//*****************************************************************************
//*
//*
//*		CFirFilter
//*
//*
//*****************************************************************************
//	Diese Klasse Filtert einen Datenstrom mittels FIR-Filter
#include	<math.h>
#include	<string.h>
#include	"FirFilter.h"

#define		PI			3.14159265358979


//*****************************************************************************
//*
//*		Konstruktor
//*
//*****************************************************************************
CFirFilter::CFirFilter()
{
	fir_len	  = 0;
	fir_coeff = 0;
	data_mem  = 0;
	data_pos  = 0;
	w_real	  = 0;
	w_count	  = 0;
}


//*****************************************************************************
//*
//*		Destruktor
//*
//*****************************************************************************
CFirFilter::~CFirFilter()
{
	if(fir_coeff)delete fir_coeff;
	if(data_mem )delete data_mem;
	if(w_real   )delete w_real;
}


//*****************************************************************************
//*
//*		Reset
//*
//*****************************************************************************
void CFirFilter::Reset()
{
	data_pos  = 0;
	memset(data_mem,0,sizeof(data_mem[0])*fir_len);
}

//*****************************************************************************
//*
//*		SetFir
//*
//*****************************************************************************
//	Stellt die FIR Koeffizienten des Filters ein
//	Coeff	: Filter Koeffizienten 
//	Count	: Anzahl der Koeffizienten 
//	Optimize: Optimiertungs Grenze
//	Ergibt die Anzahl der Koeffizienten nach der Optimierung
int	CFirFilter::SetFir(double *Coeff,int Count,double Optimize)
{
int     *fir_new ,*fir_old;
short   *data_new,*data_old;



	if(Optimize)				// Entferne zukleine Elemente
		{
		double	max;

		while(Count>=2)			// Am Anfang entfernen
			{
			   max=Coeff[0];
			if(max<0.0)max=-max;
			if(max>Optimize)break;

			Coeff++;
			Count--;
			}

		while(Count>=2)			// Am Ende entfernen
			{
			   max=Coeff[Count-1];
			if(max<0.0)max=-max;
			if(max>Optimize)break;

			Count--;
			}

		}

	

	if(fir_len == Count)		// Keine Grennderung
		{
		for(int x=0;x<Count;x++)
			{
			fir_coeff[x] = (int)(Coeff[x]*1024.0);
			}

		memset(data_mem,0,sizeof(data_mem[0])*Count);
		data_pos=0;
		}
	else{
		fir_new	  =	new int  [Count];
		data_new  =	new short[Count];
		fir_old   = fir_coeff;
		data_old  = data_mem;

		
		memset(data_new,0,sizeof(data_new[0])*Count);
		for(int x=0;x<Count;x++)
			{
			fir_new[x] = (int)(Coeff[x]*1024.0);
			}

		if(fir_len >= Count)fir_len = Count;

		fir_coeff = fir_new;
		data_mem  = data_new;
		fir_len   = Count;
		data_pos  = 0;

		if(fir_old )delete fir_old;
		if(data_old)delete data_old;
		}	


return Count;
}

//*****************************************************************************
//*
//*		Filter
//*
//*****************************************************************************
//	Stellt die FIR Koeffizienten des Filters ein
//	Output	: Speicher fr Ausgangs Samplewerte
//	Input	: Speicher fr Eingangs Samplewerte
//	Count	: Anzahl der Samplewerte
//	Channels: Anzahl der Kanle je Sample
void CFirFilter::Filter(short *Output,const short *Input,int Count,int Channels)

{
int		i,j,k,value;

	
	Count*=Channels;
	Channels--;

	for(i=0;i<Count;i++)
		{
		   data_mem[data_pos] = Input[i];		// Daten Speicher
		   data_pos++;
		if(data_pos>=fir_len)data_pos=0;										   
		k =data_pos;

		value=0;

		for(j=fir_len-1;j>=0;j--,k++)			// Faltung
			{
			if(k>=fir_len)k=0;
			value += data_mem[k]*fir_coeff[j];
			}

		   value >>= 10;						
		if(value> 32767)value =32767;			// Bereichsberlauf
		if(value<-32768)value=-32768;
		
		Output[i] = value;
		i+=Channels;
		}
}

//*****************************************************************************
//*
//*		Idft
//*
//*****************************************************************************
//	Inverse Fourier Transformation fr Reale Frequenzgnge
//	S_real	= Ausgangsspeicher fr die Impulsantwort
//	X_real  = Frequenzgang, F von X_real[i] = i/(T*Count)
//			  Es muss im Frequenzgang die Spiegelfrequenz bei 1/2T forhanden sein
//		      Ansonsten wird die Impulsantwort komplex
//	Count	= Anzahl der Frequenzsamples
void CFirFilter::Idft(double *S_real,const double *X_real,int Count)
{
	int				i, j, k;
	double			t_real;



	if(Count != w_count)				// Sin-Cos Werte erzeugen
		{
		if(w_real   )delete w_real;
		   w_real = new double[Count];
		   w_count= Count;

		for (i=0; i<Count; i++) 
			{
			w_real[i] =  cos(2.0*PI*i/(double)Count);
			}
		}

	for(k=0;k<Count;k++)
		{
		t_real=0;
		for(i=0;i<Count;i++)
			{
			j=(k*i) % Count;
			t_real += X_real[i]*w_real[j];
			}

		S_real[k]=t_real;
		}	

	
}

//*****************************************************************************
//*
//*		SetFreq
//*
//*****************************************************************************
//	Stellt die FIR Koeffizienten des Filters ein ber den Frequenzgang
//	Freq 	: Ein Feld fr den Frequenzgang von 0 Hz bis 1/2T Hz
//	Count	: Anzahl der Samplewerte 
//	Ergibt die Anzahl der Koeffizienten des FIR-Filters nach der Optimierung
int CFirFilter::SetFreq(double *Freq ,int Count,int Max)
{
double		*frq,*impuls,value;
int			 x,y;


	value = 0.5/(double)Count;
	frq = new double[Count*2];				// Spiegelfrequenz erzeugen
	memcpy(frq,Freq,sizeof(double)*Count);
	for(x=0,y=Count*2-1;x<Count;x++,y--)
		{
		frq[x] *= value;
		frq[y]  = frq[x];
		}

	impuls = new double[Count*2];
	Idft(impuls,frq,Count*2);
	delete frq;

	for(x=0;x<Count;x++)					// Impulseantwort um sortieren
		{
		value =	impuls[x];		
		impuls[x] = impuls[x+Count];
		impuls[x+Count] = value;
		}

	x=SetFir(impuls,Count*2,1.0/256.0);
	if(fir_len>Max)fir_len=Max;

	delete impuls;

return x;
}



